export const baseURL = `ws://${document.location.hostname}:443/skribbl`;
/**
 * @typedef {{connection:RTCPeerConnection,dataChannel:RTCDataChannel}} ConnectionData 
 */
/**
 * A connection to the signaling server used to start and join games.
 */
export default class Signaler {

	/**
	 * Starts a new game, returning a promise with its ID. After that, calls the given callback with a new RTCPeerConnection whenever someone joins the game.
	 * @param {(connectionData:ConnectionData)=>void} onConnect
	 */
	static async host(onConnect){
		const socket = new CustomSocket(baseURL+"/host");
		//await socket.waitUntilReady();
		let response = socket.next(message=>{
			console.log("possible response:",message);
			return message.startsWith("started ");
		});
		let id = (await response).split(" ")[1];
		socket.onMessage(async message=>{
			if (message.startsWith("join ")){
				let playerID = message.split(" ",2)[1];
				console.log("new player joined: ",playerID);
				/** @type {((message:string)=>void)[]} */
				let callbacks = [];
				socket.onMessage(message=>{
					if (message.startsWith("message "+playerID+" ")){
						let data = message.split(" ").slice(2).join(" ");
						console.log("message from player "+playerID+": ",data);
						callbacks.forEach(callback=>callback(data));
					}
				});
				onConnect(await perfectNegotiation(callback=>callbacks.push(callback),message=>{
					console.log("message to player "+playerID+": ",message);
					socket.send("message "+playerID+" "+message);
				},false));
			}
		});
		return id;
	}

	/**
	 * Joins a new game, returning an RTCPeerConnection to its host.
	 * @param {string} id
	 */
	static async join(id){
		const socket = new CustomSocket(baseURL+"/join/"+id);
		await socket.waitUntilReady();
		/** @type {((message:string)=>void)[]} */
		let callbacks = [];
		socket.onMessage(message=>{
			console.log("new message: ",message);
			callbacks.forEach(callback=>callback(message.split(" ").slice(1).join(" ")));
		});
		return await perfectNegotiation(callback=>callbacks.push(callback),message=>{
			console.log("send message:",message);
			socket.send("message "+message);
		},true);
	}
}
/** 
 * An implementation of "perfect negotiaton" as described [here](https://developer.mozilla.org/en-US/docs/Web/API/WebRTC_API/Perfect_negotiation).
 * @param {(callback:(message:string)=>void)=>void} onMessage
 * @param {(message:string)=>void} sendMessage
 * @param {boolean} polite
 * @returns {Promise<ConnectionData>}
 */
export async function perfectNegotiation(onMessage,sendMessage,polite){
	const connection = new RTCPeerConnection();
	const dataChannel = connection.createDataChannel("channel",{id:0,negotiated:true});
	let makingOffer = false;
	connection.onnegotiationneeded = async()=>{
		makingOffer = true;
		// @ts-ignore
		await connection.setLocalDescription();
		//console.log("local description:",connection.localDescription)
		sendMessage(JSON.stringify({description:connection.localDescription}));
		makingOffer = false;
	};
	connection.onicecandidate = ({candidate})=>{
		//console.log("candidate:",candidate);
		sendMessage(JSON.stringify({candidate}));
	};
	let ignoreOffer = false;
	onMessage(async message=>{
		console.log("message to negotiator:",message);
		/** @type {{description:RTCSessionDescriptionInit,candidate:undefined}|{candidate:RTCIceCandidateInit,description:undefined}} */
		let data = JSON.parse(message);
		if (data.description){
			let description = data.description;
			ignoreOffer = !polite&&(description.type=="offer"&&(makingOffer||connection.signalingState!="stable"));
			if(!ignoreOffer){
				await connection.setRemoteDescription(description);
				if (description.type=="offer"){
					// @ts-ignore
					await connection.setLocalDescription();
					sendMessage(JSON.stringify({description:connection.localDescription}));
				}
			}
		}else if(data.candidate){
			try {
				await connection.addIceCandidate(data.candidate);
			}catch(e){
				if (!ignoreOffer){
					throw e;
				}
			}
		}
	});
	/** @type {Promise<ConnectionData>} */
	return new Promise(resolve=>{
		connection.onsignalingstatechange = ()=>{
			console.log("Signaling state changed to:",connection.signalingState);
			if (connection.signalingState=="stable"){
				resolve({connection,dataChannel});
			}
		};
	});
}
/**
 * Custom socket class to pass strings to and from a server.
 * 
 * Originally, the idea was to let this pass json data instead, but since json is difficult to work with in java I've stuck with normal strings for now.
 */
export class CustomSocket {
	/**
	 * @callback MessageCallback
	 * @param {string} message
	 * @param {()=>void} remove removes this callback again
	 * @returns {void}
	 */
	/**
	 * @callback ErrorCallback
	 * @param {Event} error
	 * @param {()=>void} remove removes this callback again
	 * @returns {void}
	 */
	/**
	 * Creates a new custom socket.
	 * @param {string} url
	 */
	constructor(url){
		/** @type {MessageCallback[]} */
		this._onMessageCallbacks = [];
		/** @type {ErrorCallback[]} */
		this._onErrorCallbacks = [];
		this._webSocket = new WebSocket(url);
		/** @type {Promise<void>} */
		this._readyPromise = new Promise((resolve,reject)=>{
			this._webSocket.onopen = ()=>resolve();
			this._webSocket.addEventListener("error",reject,{once:true});
		});
		this._webSocket.onmessage = (e)=>{
			console.log(`onmessage event! e.data: "${e.data}", registered callbacks: ${this._onMessageCallbacks.length}`);
			this._onMessageCallbacks.slice().forEach(callback=>callback(e.data+"",()=>this.removeMessageCallback(callback)));
		};
		this._webSocket.onerror = (e)=>{
			this._onErrorCallbacks.slice().forEach(callback=>callback(e,()=>this.removeErrorCallback(callback)));
		};
	}

	async waitUntilReady(){
		return this._readyPromise;
	}

	/**
	 * Sends the given object as json to the server.
	 * @param {string} data
	 */
	send(data){
		this._webSocket.send(data);
	}

	/**
	 * Registers a callback to be called whenever a message is received.
	 * @param {MessageCallback} callback
	 */
	onMessage(callback){
		this._onMessageCallbacks.push(callback);
	}

	/**
	 * Removes a callback registered using `onMessage(callback)`.
	 * @param {MessageCallback} callback
	 */
	removeMessageCallback(callback){
		let index = this._onMessageCallbacks.indexOf(callback);
		if (index!==-1){
			this._onMessageCallbacks.splice(index,1);
		}
	}

	/**
	 * Returns a Promise that resolves to the next message received that passen the given filter, or `null` if no such message is received in the specified timeframe.
	 * If an error occurs before, the promise will instead be rejected with that.
	 * @param {(message:string)=>boolean} filter
	 * @param {number} time time to wait before just returning `null`, in seconds. Defaults to 30. If the given number is smaller than or equal to 0, it gets treated as `Infinity`.
	 * @return {Promise<string>}
	 */
	async next(filter=null,time=30){
		return new Promise((resolve,reject)=>{
			this.onMessage((message,remove)=>{
				if (!filter||filter(message)){
					resolve(message);
					remove();
				}
			});
			this.onError((error,remove)=>{
				reject(error);
				remove();
			});
			if (time>0){
				setTimeout(()=>resolve(null),time*1000);
			}
		});
	}

	/**
	 * Registers a callback to be called whenever there's an error.
	 * @param {ErrorCallback} callback
	 */
	onError(callback){
		this._onErrorCallbacks.push(callback);
	}

	/**
	 * Removes a callback registered using `onError(callback)`.
	 * @param {ErrorCallback} callback
	 */
	removeErrorCallback(callback){
		let index = this._onErrorCallbacks.indexOf(callback);
		if (index!==-1){
			this._onErrorCallbacks.splice(index,1);
		}
	}

	/**
	 * Closes the connection.
	 */
	close(){
		this._webSocket.close();
	}
}